from typing import Any, Union

import requests
from openai import ChatCompletion

from ..models.chat_format import Message
from ..util import get_dict
from .metadata import generate_metadata_dict


class openai:
    @staticmethod
    def log(
        response: ChatCompletion,
        model: str,
        messages: list[dict[str, str]],
        session_key: str = None,
        **kwargs,
    ) -> dict[str, str]:
        body: dict[str, str] = {
            "input": {
                "model": model,
                "messages": messages,
                **kwargs,
            },
            "output": get_dict(response),
            "message": get_dict(response)["choices"][0]["message"]["content"],
            "metadata": generate_metadata_dict(),
        }
        request: requests.Response = requests.post(
            f"{ENDPOINT}/logging/create/llm", json=body
        )
        return request.json()

    @staticmethod
    def log_verbose(
        model: str,
        messages: Union[list[Message], list[dict[str, str]]],
        completion: dict[str, str],
        session_key: str = None,
        **kwargs,
    ) -> dict[str, str]:
        body: dict[str, str] = {
            "input": {
                "model": model,
                "messages": messages,
                **kwargs,
            },
            "output": completion,
            "message": completion["choices"][0]["message"]["content"],
            "metadata": generate_metadata_dict(),
        }
        request: requests.Response = requests.post(
            f"{ENDPOINT}/logging/create/llm", json=body
        )
        print(request.json())
        return request.json()



import requests

from .metadata import generate_metadata_dict


class app:
    def log(message: str, endpoint: str = "https://api.speck.chat") -> dict[str, str]:
        body: dict[str, str] = {
            "message": message,
            "metadata": generate_metadata_dict(),
        }
        request: requests.Response = requests.post(
            f"{endpoint}/logging/create/app", json=body
        )
        return request.json()



def db_write_format():
    return {
        "type": "llm",
        "userId": "test",
        "message": "test",
        "timestamp": "test",
        "input": "test",
        "output": "test",
    }


def db_read_format():
    return {
        "type": "llm",
        "userId": "test",
        "message": "test",
        "timestamp": "test",
        "input": "test",
        "output": "test",
    }



from abc import ABC

from ..chat.entities import ChatLogger, Prompt, Response
from ..logs.logger import universal_format_log
from .providers import Providers


class IConnector(ABC):
    _client: "Speck"

    def __init__(self, client: "Speck", provider: Providers):
        self._client = client
        self.provider = provider

    # @abstractmethod
    # def process_message(self, messages: Messages, model: str) -> str:
    #     pass

    def _get_log_kwargs(self, prompt: Prompt, response: Response, **kwargs):
        return {
            "provider": self.provider,
            "model": kwargs.get("model"),
            "temperature": kwargs.get("temperature"),
            "stream": kwargs.get("stream", False),
            "prompt": prompt,
            "config": kwargs,
            "response": response,
        }

    def log(self, *, prompt: Prompt, response: Response, **kwargs):
        ChatLogger.log(
            self._client.endpoint, **self._get_log_kwargs(prompt, response, **kwargs)
        )

    def __str__(self):
        return f"Client({self.provider.value})"



from typing import Union

from .chat.entities import ChatConfig, Prompt, Response
from .connections.anthropic import AnthropicConnector
from .connections.openai import OpenAIConnector
from .connections.openai_azure import AzureOpenAIConnector
from .connections.replicate import ReplicateConnector
from .logs.app import app


# Todo: Add BaseClient
# Todo: Add AsyncResource, SyncResource
class BaseClient:
    pass


class Resource:
    pass


class AsyncResource(Resource):
    pass


class SyncResource(Resource):
    pass


class Logger(SyncResource):  # App logger
    def __init__(self, client: "Speck"):
        self.client = client

    def log(self, *args, **kwargs):
        kwargs["endpoint"] = self.client.endpoint
        app.log(*args, **kwargs)


class Chat(SyncResource):
    def __init__(self, client: "Speck"):
        self.client = client

    def create(self, *, prompt: Prompt, config: ChatConfig = None, **config_kwargs):
        if config is None:
            config = ChatConfig(**config_kwargs)
            # Todo: convert to default config based on class param

        if config.provider is None:
            # Try to extract provider by getting string before : in model
            if ":" in config.model:
                provider_str, model_str = config.model.split(":", 1)
                config.provider = provider_str
                config.model = model_str
            else:
                raise ValueError(
                    "Provider must be specified in config or as a class param"
                )

        if config.provider == "openai":
            connector = OpenAIConnector(
                client=self.client, api_key=self.client.api_keys["openai"].strip()
            )
            return connector.chat(prompt, config, **config_kwargs)
        if config.provider == "azure-openai":
            connector = AzureOpenAIConnector(
                client=self.client,
                api_key=self.client.api_keys["azure-openai"].strip(),
                **self.client.azure_openai_config
            )
            return connector.chat(prompt, config, **config_kwargs)
        if config.provider == "replicate":
            connector = ReplicateConnector(
                client=self.client, api_key=self.client.api_keys["replicate"].strip()
            )
            return connector.chat(prompt, config, **config_kwargs)
        if config.provider == "anthropic":
            connector = AnthropicConnector(
                client=self.client, api_key=self.client.api_keys["anthropic"].strip()
            )
            return connector.chat(prompt, config, **config_kwargs)
        raise ValueError("Provider not found")

    def log(self, messages: Prompt, config: ChatConfig, response: Response):
        config.log_chat(
            endpoint=self.client.endpoint, prompt=messages, response=response
        )


class AsyncChat(AsyncResource):
    def __init__(self, client: "AsyncSpeck"):
        self.client = client

    def log(self, messages: Prompt, config: ChatConfig, response: Response):
        config.log_chat(
            endpoint=self.client.endpoint, prompt=messages, response=response
        )


class Speck(BaseClient):
    api_key: Union[str, None]
    api_keys: dict[str, str]
    endpoint: str

    def __init__(
        self,
        api_key: Union[str, None] = None,
        api_keys: dict[str, str] = {},
        endpoint: str = "https://api.speck.chat",
    ):
        self.api_key = api_key
        self.api_keys = api_keys
        self.endpoint = endpoint

        self.azure_openai_config = {}

        self.chat = Chat(self)
        self.logger = Logger(self)

    def add_api_key(self, provider: str, api_key: str):
        self.api_keys[provider] = api_key

    def add_azure_openai_config(self, azure_endpoint: str, api_version: str):
        self.azure_openai_config = {
            "azure_endpoint": azure_endpoint,
            "api_version": api_version,
        }


class AsyncSpeck(BaseClient):
    api_key: Union[str, None] = None

    def __init__(self, api_key: Union[str, None] = None):
        self.api_key = api_key
        self.chat = AsyncChat(self)


Client = Speck
AsyncClient = AsyncSpeck



"""
Name: OpenAI Azure
URL:
Features:
"""
from openai import AzureOpenAI

from .openai import OpenAIConnector


class AzureOpenAIConnector(OpenAIConnector):
    def __init__(self, api_key: str, azure_endpoint: str, api_version: str, **kwargs):
        super().__init__(api_key=None)
        self.client = AzureOpenAI(
            api_key=self.api_key,
            azure_endpoint=azure_endpoint,
            api_version=api_version,
            **kwargs
        )
        self.api_key = api_key
        self.azure_endpoint = azure_endpoint
        self.api_version = api_version



import json
import logging
import os
from typing import Union

import httpx
import requests
from openai import OpenAI

from ..logs import logger
from .chat_format import Message

client = OpenAI(api_key="hi")


def get_api_key():
    return os.environ["OPENAI_API_KEY"]


# Todo: migrate this to its own chat folder
class chat:
    @staticmethod
    def create(
        model: str,
        messages: Union[list[Message], list[dict[str, str]]],
        session_key: str = None,
        local_retry: bool = True,
        **kwargs,
    ):
        try:
            body: dict[str, str] = {
                "model": model,
                "messages": messages,
            }
            request: requests.Response = requests.post(
                f"{ENDPOINT}/completions/openai/create", json=body
            )
            request.raise_for_status()
            return request.json()
        except requests.exceptions.HTTPError as e:
            if local_retry and get_api_key() is not None:
                logging.warning(
                    "Failed to create completions, retrying with local API key"
                )
                return client.chat.completions.create(
                    messages=messages,
                    model=model,
                )
            else:
                raise e
        except requests.exceptions.ReadTimeout as errrt:
            logging.warning("Time out")
        except requests.exceptions.ConnectionError as conerr:
            logging.warning("Connection error")
        except requests.exceptions.RequestException as errex:
            logging.warning("Exception request")

    @staticmethod
    async def create_async(
        model: str,
        messages: Union[list[Message], list[dict[str, str]]],
        session_key: str,
        **kwargs,
    ):
        # TODO: post async
        pass

    # Convert streaming to the following format:
    # for text, chunk in chat.stream(model, messages, session_key):
    @staticmethod
    def create_stream(
        model: str,
        messages: Union[list[Message], list[dict[str, str]]],
        process_chunk_lambda=None,
        session_key: str = None,
        log: bool = False,
    ):
        body: dict[str, str] = {
            "model": model,
            "messages": messages,
        }

        with httpx.Client() as client:
            with client.stream(
                "POST", f"{ENDPOINT}/completions/openai/stream", json=body, timeout=None
            ) as response:
                for chunk in response.iter_raw():
                    decoded_chunk: str = chunk.decode("utf-8")
                    chunk_data: dict[str, str] = json.loads(decoded_chunk)
                    chunk_type: str = chunk_data.get("type")
                    chunk_content: str = chunk_data.get("content")

                    # Process each chunk with the lambda function
                    process_chunk_lambda(chunk_content, chunk_type == "full")

                    # Handle the end of the stream
                    if chunk_type == "full":
                        return chunk_content



from ..chat.entities import IChatClient, Prompt, Response
from .connector import IConnector
from .providers import Providers


class CustomProviderConnector(IConnector, IChatClient):
    def __init__(self, client: "Speck", message_prefix: str, message_suffix: str):
        super().__init__(client=client, provider=Providers.CustomProvider)
        self.message_prefix = message_prefix
        self.message_suffix = message_suffix

    def chat(self, prompt: Prompt, model: str, **config_kwargs) -> Response:
        return Response(f"{self.message_prefix} {prompt} {self.message_suffix}")



import logging
from typing import Any, Callable, Iterator

from openai import OpenAI as _OpenAI
from openai import Stream as _Stream

from .. import ChatLogger
from ..chat.entities import ChatConfig, OpenAIChatConfig, Prompt, Response, Stream
from ..util._wrapper import wrap_method

logger = logging.getLogger(__name__)
logger.info = print


def wrapper(original_method, *args, **kwargs):
    """
    Example of a wrapper function that can be used with add_method_wrapper.

    Args:
        original_method: The original method to be wrapped.
        *args, **kwargs: Arguments and keyword arguments for the original method.

    Returns:
        The result of the original method call.
    """
    logger.info(f"Before calling {original_method.__name__}")
    result = original_method(*args, **kwargs)
    logger.info(f"After calling {original_method.__name__}")
    return result


class OpenAIStream:
    # processor that has lambda which returns MessageDelta
    def __init__(
        self,
        config: OpenAIChatConfig,
        prompt: Prompt,
        iterator: Iterator[Any],
        kwargs: dict,
        processor: Callable[[Any], Any],
    ):
        self.config: OpenAIChatConfig = config
        self.prompt = prompt
        self.message: str = ""
        self._iterator = iterator
        self._kwargs = kwargs
        self._processor = processor
        self._has_logged = False
        self._closed = False
        self._last_item = None

    def _log(self):
        if not self._has_logged:
            self._has_logged = True

            kwargs = self._kwargs
            kwargs["prompt"] = self._kwargs.get("prompt", [])
            kwargs["model"] = self._kwargs.get("model", "N/A")
            # kwargs["response"] = Response(content=self.message, raw={}, closed=True)
            # ChatLogger.log(**kwargs)

            print("Logging chat!", self.message)
            self.config.log_chat(
                endpoint="https://api.speck.chat",
                prompt=self.prompt,
                response=Response(content=self.message, raw={}, closed=True),
            )

    def _process(self, item) -> Any:
        return self._processor(item)

    def __next__(self) -> Any:
        try:
            if self._closed:
                raise StopIteration

            item: Any = self._process(next(self._iterator))
            self._last_item = item
            if item.choices[0].delta.content:
                self.message += item.choices[0].delta.content
            return item
        except StopIteration:
            self._log()
            raise

    def __iter__(self) -> Iterator[Any]:
        return self

    def close(self):
        try:
            self._closed = True
            # todo: make this work for packages other than openai
            self._iterator.response.close()
        except AttributeError:
            pass


def chat_wrapper(original_method, *args, **kwargs):
    """
    Example of a wrapper function that can be used with add_method_wrapper.

    Args:
        original_method: The original method to be wrapped.
        *args, **kwargs: Arguments and keyword arguments for the original method.

    Returns:
        The result of the original method call.
    """
    model = kwargs.get("model", None)
    stream = kwargs.get("stream", False)
    messages = kwargs.get("messages", None)
    prompt = Prompt.from_openai(messages)

    config = OpenAIChatConfig(**kwargs)

    if model is not None and ":" in model:
        model = model.split(":")[1]
        # Todo: convert other model responses to OpenAI response format

    logger.info(f"Call {original_method.__name__} with model {model}")
    if stream:
        print("Stream called!")
        stream = original_method(*args, **kwargs)
        # Todo: wrap the Stream class
        # Best current way to do this is to convert our Stream class to an OpenAI Stream class
        return OpenAIStream(
            config=config,
            prompt=prompt,
            iterator=stream,
            kwargs={
                "provider": "openai",
            },
            processor=lambda a: a,
        )
    else:
        result = original_method(*args, **kwargs)

        config.log_chat(
            endpoint="https://api.speck.chat",
            prompt=prompt,
            response=result,
            provider="openai",
        )

        return result


def stream_next_wrapper(original_method, *args, **kwargs):
    print(f"Before calling {original_method.__name__}")
    result = original_method(*args, **kwargs)
    print(f"After calling {original_method.__name__}")
    print(result)
    return result


def stream_iter_wrapper(original_method, *args, **kwargs):
    print(f"Before calling {original_method.__name__}")
    result = original_method(*args, **kwargs)
    print(f"After calling {original_method.__name__}")
    print(result)
    return result


class OpenAIWrapper(_OpenAI):
    pass


def _wrapper_init(original_method, *args, **kwargs):
    """
    Example of a wrapper function that can be used with add_method_wrapper.

    Args:
        original_method: The original method to be wrapped.
        *args, **kwargs: Arguments and keyword arguments for the original method.

    Returns:
        The result of the original method call.
    """
    logger.info(f"Initializing {original_method.__name__}")
    result = original_method(*args, **kwargs)
    logger.info(f"Adding method wrappers {original_method.__name__}")
    self = args[0]
    wrap_method(self.chat.completions, "create", chat_wrapper)
    logger.info(f"After calling {original_method.__name__}")
    return result


wrap_method(OpenAIWrapper, "__init__", _wrapper_init)

# add_method_kwarg(OpenAIWrapper, "__init__", "speck_log", 69)



class Promptfile:
    """A class for loading and parsing .prompt files"""

    def __init__(self, filepath: str):
        self.filepath = filepath
        self.promptfile = self._load_promptfile()

    def _load_promptfile(self) -> dict:
        """Load the promptfile from the filepath."""
        return {}

    def _parse_promptfile(self) -> dict:
        """Parse the promptfile."""
        return {}
